/* KallistiOS Ogg/Vorbis Decoder Library
 *
 * sndvorbisfile.c
 * (c)2001/2002 Thorsten Titze
 *
 * Basic Ogg/Vorbis stream information and decoding routines used by
 * libsndoggvorbis. May also be used directly for playback without
 * threading.
 */

static char id[] = "sndoggvorbis $Id: sndvorbisfile.c,v 1.6 2002/04/01 02:31:02 bardtx Exp $";
#include <kos.h>
#include <stdio.h>
#include <vorbis/codec.h>
#include "sndvorbisfile.h"

VorbisFile_handle_t fd;

// Ogg/Vorbis relevant variables
ogg_sync_state		oy;
ogg_page		og;
ogg_packet		op;
ogg_stream_state	os;
vorbis_info		vi;
vorbis_comment		vc;
vorbis_dsp_state	vd;
vorbis_block		vb;

int			VorbisFile_convbufferlength=4096;
int			VorbisFile_EOS;

/* Variables for calculation of actual bitrate */
double			VorbisFile_sampletrack;
double			VorbisFile_bittrack;

char			*VorbisFile_streambuffer;
int			bytes;
uint32			fw;

/* Defines for various VorbisFile status */
#define VorbisFile_EOSreached	1
#define VorbisFile_notplaying	-1
#define VorbisFile_playing	1

VorbisFile_handle_t fd;

/* VorbisFile_isEOS()
 *
 * detects whether or not we already reached the end of our Ogg/Vorbis stream
 *
 * Returns:
 * 1  = Reached End Of Stream
 * 0  = Well.. I guess it's clear.. We did not yet reach the End Of Stream
 */
int VorbisFile_isEOS()
{
	if (VorbisFile_EOS==1)return(1);
	else return(0);
}

/* VorbisFile_getBitrateNominal()
 *
 * returns the nominal bitrate of the already loaded bitstream.
 *
 * Returns:
 * LONG = Nominal bitrate of the loaded Strea
 * -1   = Clear also: No stream loaded yet.
 */
long VorbisFile_getBitrateNominal()
{
	if (fd == 0)return(-1);
	else return(vi.bitrate_nominal);
}

/* VorbisFile_getBitrateInstant()
 *
 * returns the actual bitrate of the file since the last call
 *
 * Returns:
 * LONG = Bitrate
 * -1   = No stream loaded.
 */
long VorbisFile_getBitrateInstant()
{
	long ret;

	if (fd == 0)return(-1);
	if (VorbisFile_sampletrack == 0)return(-1); /* prevent division by zero */

	/* Calcualte actual bitrate */
	ret=VorbisFile_bittrack/VorbisFile_sampletrack*vi.rate+.5;

	/* reset tracking variables for next call */
	VorbisFile_bittrack = 0.0f;
	VorbisFile_sampletrack = 0.0f;

	return(ret);
}

/* VorbisFile_getCommentByName(...)
 *
 * returns the comment value of a specified commentfield
 *
 * Returns:
 * Success = comment content
 * NULL = field not found
 */
char *VorbisFile_getCommentByName(char *commentfield)
{
	int i;
	int commentlength=strlen(commentfield);
	
	if (fd == 0) return(VorbisFile_NULL);
	
	//return(vorbis_comment_query(&vc, commentfield,1));
	for(i=0;i<vc.comments;i++)
	{
		if(!(strncmp(commentfield,vc.user_comments[i],commentlength)))
		{
			/* Return adress of comment content but leave out the
			 * commentname= part !
			 */
			return(vc.user_comments[i]+commentlength+1);
		}
	}
	return(VorbisFile_NULL);
}

/* VorbisFile_getCommentByID(...)
 *
 * returns the comment identified by id
 *
 * Returns:
 * Success = comment content
 * NULL = field not found
 */
char *VorbisFile_getCommentByID(long commentid)
{
	if (fd == 0)return(NULL);

	/* Check if we have at least that much comments in our file ?! */
	if (vc.comments >0)
	{
		return vc.user_comments[commentid];
	}
	return(VorbisFile_NULL);
}
	
/* VorbisFile_closeFile()
 *
 * Closes the Filehandle used for accessing the OggVorbis-Stream
 */
void VorbisFile_closeFile()
{
	if (fd != 0)
	{
		fs_close(fd);
		vorbis_block_clear(&vb);
		vorbis_dsp_clear(&vd);
		ogg_stream_clear(&os);
		vorbis_info_clear(&vi);
		vorbis_comment_clear(&vc);
		ogg_sync_clear(&oy);
	}
}

/* int VorbisFile_openFile(...)
 *
 * returns:
 *  0 = File open and header read succesful
 * -1 = File not opened successfully or no valid OggVorbis stream
 */
int VorbisFile_openFile(char *filename, VorbisFile_headers_t *v_headers)
{
	int	i;
	VorbisFile_EOS = 0;

	printf("liboggvorbis: trying to open file \"%s\"\r\n",filename);

	fd= fs_open(filename, O_RDONLY);

	if (fd == 0)
	{
		printf("liboggvorbis: cannot open file \"%s\"\r\n",filename);
		return (-1);
	}

	ogg_sync_init(&oy);
	
	VorbisFile_streambuffer=ogg_sync_buffer(&oy,4096);
	bytes = fs_read(fd, VorbisFile_streambuffer, 4096);
	
	ogg_sync_wrote(&oy,bytes);

	if(ogg_sync_pageout(&oy,&og)!=1)
	{
		if(bytes<4096)return (-1);
	}
	printf("libogg: input bitstream has been detected to be Ogg compliant\n");
	
	ogg_stream_init(&os,ogg_page_serialno(&og));
	printf("libogg: ogg input bitstream initialized\n");
	
	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);

	printf("libvorbis: info and comment initialized\n");
	
	if(ogg_stream_pagein(&os,&og)<0)
	{
		printf("libogg: error reading first page of Ogg bitstream\n");
	}

	if(ogg_stream_packetout(&os,&op)!=1)
	{
		printf("libogg: error reading initial header packet\n");
	}

	if(vorbis_synthesis_headerin(&vi,&vc,&op)<0)
	{
		printf("libvorbis: Ogg bitstream does not contain Vorbis encoded date\n");
	}
	
		
	i=0;
	while(i<2)
	{
		while(i<2)
		{
			int result=ogg_sync_pageout(&oy,&og);
			if(result==0)break;
			if(result==1)
			{
				ogg_stream_pagein(&os,&og);
				while(i<2)
				{
					result=ogg_stream_packetout(&os,&op);
					if(result==0)break;
					if(result<0)
					{
						printf("libogg: corrupt secondary header in bitstream\n");
					}
					vorbis_synthesis_headerin(&vi,&vc,&op);
					i++;
				}
			}
		}
		VorbisFile_streambuffer=ogg_sync_buffer(&oy,4096);
		bytes = fs_read(fd, VorbisFile_streambuffer, 4096);
		if(bytes==0 && i<2)
		{
			printf("libogg: EOF occured before all headers have been found\n");
		}
		ogg_sync_wrote(&oy,bytes);
	}

	printf("libvorbis: Bitstream is %d channel, %ld Hz, %ld bit/s\r\n",vi.channels,vi.rate,vi.bitrate_nominal);
	printf("libvorbis: Encoded by: %s\r\n",vc.vendor);
	printf("libvorbis: Found %d additional comment fields\r\n",vc.comments);

	/* Todo:
	 * Reading of additional comment fields and putting in v_headers
	 */

	v_headers->channels=vi.channels;
	v_headers->frequency=vi.rate;
	v_headers->bitrate=vi.bitrate_nominal;
	v_headers->vendor=vc.vendor;
	v_headers->convsize=VorbisFile_convbufferlength/vi.channels;

	/* Initialize bitrate tracking variables */
	VorbisFile_bittrack = 0;
	VorbisFile_sampletrack = 0;

	// Initialize the decoder
	vorbis_synthesis_init(&vd,&vi);
	vorbis_block_init(&vd,&vb);

	return (0);
}

/* int VorbisFile_decodePCM(...)
 *
 * dependent on the target-buffer size we have to change to pointer arithmetic to
 * reflect other datatypes. this example as it is coded here reflects an int16
 * data-type for PCM buffer storage.
 * 
 * returns:
 * >0 = number of bytes decoded
 * -1 = we have reached the end of the Ogg stream
 */
// Subroutine that grabs stream data from a file
//
void iVorbisFile_grabData()
{
	VorbisFile_streambuffer=ogg_sync_buffer(&oy,4096);
	bytes = fs_read(fd, VorbisFile_streambuffer, 4096);
	ogg_sync_wrote(&oy,bytes);
}

int VorbisFile_decodePCMint8(VorbisFile_headers_t vhd, uint8 *target, int requested)
{
	int convsize=vhd.convsize;
	int i;

	int haveenough=0;
	int wehave=0;

	while (!haveenough)
	{
		float	**pcm;
		int	samples;

		// first we have to try to get some data out of the current stream
		while((samples=vorbis_synthesis_pcmout(&vd,&pcm))>0)
		{
			VorbisFile_sampletrack += samples;
			int j;
			// int clipflag=0;

			// the "bout" variable holds the number of PCM samples we got. but this
			// only counts how many stereo-16bit samples we have.
			//
			// so to get the number of 16-bit samples we have to multiply by 2
			// and to get the number of bytes at all we have to multiply by 4
			int bout=(samples<convsize?samples:convsize);
			int numberofbyte;

			// pcm contains both (left and right) decoded pcm values
			for(i=0;i<vi.channels;i++)
			{
				uint8 *ptr=target+(i*2);

				// *mono will hold informations for only one channel
				float	*mono=pcm[i];
				for(j=0;j<bout;j++)
				{
					int val=mono[j]*32767.f;
						
					// Check for clipping
					if(val>32767)
					{
						val=32767;
						// clipflag=1;
					}
					if(val<-32768)
					{
						val=-32768;
						// clipflag=1;
					}
					// First store first eight bits
					*ptr= val;
					ptr+=1;
					// Now we store second eight bits
					*ptr= val >> 8;
					ptr+=3;
				}
			}
			
			/* Clipping debug output are no longer interesting !
			if(clipflag)
			{
			  printf("liboggvorbis: clipping occured. continuing\n");
			}
			*/
			
			// *2 = number of integer samples  * 2 = number of bytes
			numberofbyte= 4 * bout;

			target+=numberofbyte;
			
			vorbis_synthesis_read(&vd,bout);
			// printf("liboggvorbis: decoded %d pcm samples\r\n", bout);

			wehave += numberofbyte;
			if (wehave >= requested)
			{
				haveenough = 1;
				break;
			}
		}

		if (!haveenough)
		{
			while(ogg_stream_packetout(&os,&op)==0)
			{
				while(ogg_sync_pageout(&oy,&og)==0)
				{
					// printf("liboggvorbis: had to grab more data from file\r\n");
					iVorbisFile_grabData();
				}

				/* If the page contains the EOS flag just get out of here
				 */
				if(ogg_page_eos(&og))
				{
					VorbisFile_EOS = 1;
					printf("liboggvorbis: reached the end of the stream\r\n");
					return(-1);
				}
				// printf("liboggvorbis: ogg_stream_pagein\r\n");
				ogg_stream_pagein(&os,&og);
			}
			if(vorbis_synthesis(&vb,&op)==0)
			{
				vorbis_synthesis_blockin(&vd,&vb);

				VorbisFile_bittrack += op.bytes * 8;
			}
		}
	}
	// printf("liboggvorbis: decoded enough data\r\n");
	return(wehave);
}

/* VorbisFile_decodePCM(...)
 * 
 * same as VorbisFile_decodePCMint8 but decodes into a 16-bit Integer buffer not in
 * a 8 bit byte buffer
 */ 
int VorbisFile_decodePCM(VorbisFile_headers_t vhd, ogg_int16_t *target, int requested)
{
	int convsize=vhd.convsize;
	int i;

	int haveenough=0;
	int wehave=0;

	while (!haveenough)
	{
		float	**pcm;
		int	samples;

		// first we have to try to get some data out of the current stream
		while((samples=vorbis_synthesis_pcmout(&vd,&pcm))>0)
		{
			int j;
			// int clipflag=0;

			// the "bout" variable holds the number of PCM samples we got. but this
			// only counts how many stereo-16bit samples we have.
			//
			// so to get the number of 16-bit samples we have to multiply by 2
			// and to get the number of bytes at all we have to multiply by 4
			int bout=(samples<convsize?samples:convsize);
			int numberofbyte;

			// pcm contains both (left and right) decoded pcm values
			for(i=0;i<vi.channels;i++)
			{
				ogg_int16_t *ptr=target+i;

				// *mono will hold informations for only one channel
				float	*mono=pcm[i];
				for(j=0;j<bout;j++)
				{
					int val=mono[j]*32767.f;
						
					// Check for clipping
					if(val>32767)
					{
						val=32767;
						//clipflag=1;
					}
					if(val<-32768)
					{
						val=-32768;
						//clipflag=1;
					}
					*ptr=val;
					ptr+=vi.channels;
				}
			}
			/*
			if(clipflag)
			{
			  printf("liboggvorbis: clipping occured. continuing\n");
			}
			*/
			numberofbyte= 4 * bout;

			// See above why we have to use *2 !!!
			target+=(bout * 2);
			
			vorbis_synthesis_read(&vd,bout);
			// printf("liboggvorbis: decoded %d pcm samples\r\n", bout);

			wehave += numberofbyte;
			if (wehave >= requested)
			{
				haveenough = 1;
				break;
			}
		}

		if (!haveenough)
		{
			while(ogg_stream_packetout(&os,&op)==0)
			{
				while(ogg_sync_pageout(&oy,&og)==0)
				{
					// printf("liboggvorbis: had to grab more data from file\r\n");
					iVorbisFile_grabData();
				}
				if(ogg_page_eos(&og))
				{
				 	VorbisFile_EOS = 1;
					printf("liboggvorbis: reached the end of the stream\r\n");
					return(-1);
				}
				// printf("liboggvorbis: ogg_stream_pagein\r\n");
				ogg_stream_pagein(&os,&og);
			}
			if(vorbis_synthesis(&vb,&op)==0)
			{
				vorbis_synthesis_blockin(&vd,&vb);
			}
		}
	}
	// printf("liboggvorbis: decoded enough data\r\n");
	return(wehave);
}

